#!/usr/bin/env python3

# https://pypi.org/project/hexdump/
import hexdump

import socket, struct, random, math
import hashlib, hmac
import sys

# https://pycryptodome.readthedocs.io/en/latest/src/installation.html
# https://pycryptodome.readthedocs.io/en/latest/src/cipher/aes.html
import Crypto.Cipher.AES
# https://pycryptodome.readthedocs.io/en/latest/src/util/util.html
import Crypto.Util.Counter

HOST=None
PORT=22
USERNAME=None
PASSWORD=None
ENCRYPTION=True
VERBOSITY=0
COMMAND="uname -a" # default

def parse_command_line():
    global HOST, PORT, USERNAME, PASSWORD, ENCRYPTION, VERBOSITY, COMMAND
    if len(sys.argv)==1:
        print ("Error: no options supplied. Use these:")
        print ("  -v, -vv, -vvv    - increase verbosity level")
        print ("  --no-encryption  - to be used with patched OpenSSH")
        print ("  -h host")
        print ("  -port port       - default is 22")
        print ("  -u user")
        print ("  -pass password")
        print ("  -c command       - run command on SSH host")
        exit(0)
    idx=1
    while idx<len(sys.argv):
        if sys.argv[idx]=="-no-encryption":
            ENCRYPTION=False
            idx=idx+1
        elif sys.argv[idx]=="-v":
            VERBOSITY=1
            idx=idx+1
        elif sys.argv[idx]=="-vv":
            VERBOSITY=2
            idx=idx+1
        elif sys.argv[idx]=="-vvv":
            VERBOSITY=3
            idx=idx+1
        elif sys.argv[idx]=="-h":
            HOST=sys.argv[idx+1]
            idx=idx+2
        elif sys.argv[idx]=="-port":
            PORT=int(sys.argv[idx+1])
            idx=idx+2
        elif sys.argv[idx]=="-u":
            USERNAME=sys.argv[idx+1]
            idx=idx+2
        elif sys.argv[idx]=="-pass":
            PASSWORD=sys.argv[idx+1]
            idx=idx+2
        elif sys.argv[idx]=="-c":
            COMMAND=sys.argv[idx+1]
            idx=idx+2
        else:
            print ("Fatal error: unknown command option:", sys.argv[idx])
            exit(0)

#CIPHER="aes128-ctr"
CIPHER="aes256-ctr"

#HASH="hmac-sha1"
HASH="hmac-sha2-256"

if HASH=="hmac-sha2-256":
    MAC_SIZE=0x20
    MAC_ALGO=hashlib.sha256 # ptr to func
elif HASH=="hmac-sha1":
    MAC_SIZE=160//8 # in bytes
    MAC_ALGO=hashlib.sha1 # ptr to func
else:
    assert False

if ENCRYPTION:
    EXT_INFO_C=False
else:
    EXT_INFO_C=True

#DH_GEX_min, DH_GEX_bits, DH_GEX_max = 2048, 8192, 8192 # slow in Python!
#DH_GEX_min, DH_GEX_bits, DH_GEX_max = 2048, 2048, 2048
DH_GEX_min, DH_GEX_bits, DH_GEX_max = 512, 512, 8192

client_banner=b'SSH-2.0-ToySSH_0.1'

SSH_MSG_DISCONNECT=1
SSH_MSG_IGNORE=2
SSH_MSG_UNIMPLEMENTED=3
SSH_MSG_SERVICE_REQUEST=5
SSH_MSG_SERVICE_ACCEPT=6
SSH2_MSG_EXT_INFO=7 # from OpenSSH 9.0

SSH_MSG_KEXINIT=20
SSH_MSG_NEWKEYS=21 # 0x15

SSH_MSG_KEXDH_REPLY=31 # 0x1f
SSH_MSG_KEXDH_INIT=30 # 0x1e
SSH_MSG_KEX_DH_GEX_INIT=32 # 0x20
SSH_MSG_KEX_DH_GEX_REPLY=33 # 0x21
SSH_MSG_KEX_DH_GEX_REQUEST=34 # 0x22

SSH_MSG_USERAUTH_REQUEST=50 # 0x32 
SSH_MSG_USERAUTH_FAILURE=51 # 0x33 
SSH_MSG_USERAUTH_SUCCESS=52 # 0x34 
SSH_MSG_USERAUTH_BANNER=53 # 0x35

SSH_MSG_GLOBAL_REQUEST=80 # 0x50
SSH_MSG_CHANNEL_OPEN=90 # 0x5a
SSH_MSG_CHANNEL_OPEN_CONFIRMATION=91 # 0x5b
SSH_MSG_CHANNEL_WINDOW_ADJUST=93 # 0x5d
SSH_MSG_CHANNEL_DATA=94 # 0x5e
SSH_MSG_CHANNEL_EXTENDED_DATA=95 # 0x5f
SSH_MSG_CHANNEL_EOF=96 # 0x60
SSH_MSG_CHANNEL_CLOSE=97 # 0x61
SSH_MSG_CHANNEL_REQUEST=98 # 0x62
SSH_MSG_CHANNEL_SUCCESS=99 # 0x63

msg_str={
    SSH_MSG_DISCONNECT                : "SSH_MSG_DISCONNECT",
    SSH_MSG_IGNORE                    : "SSH_MSG_IGNORE",
    SSH_MSG_UNIMPLEMENTED             : "SSH_MSG_UNIMPLEMENTED",
    SSH_MSG_SERVICE_REQUEST           : "SSH_MSG_SERVICE_REQUEST",
    SSH_MSG_SERVICE_ACCEPT            : "SSH_MSG_SERVICE_ACCEPT",
    SSH2_MSG_EXT_INFO                 : "SSH2_MSG_EXT_INFO",
    SSH_MSG_KEXINIT                   : "SSH_MSG_KEXINIT",
    SSH_MSG_NEWKEYS                   : "SSH_MSG_NEWKEYS",
    SSH_MSG_KEXDH_REPLY               : "SSH_MSG_KEXDH_REPLY",
    SSH_MSG_KEXDH_INIT                : "SSH_MSG_KEXDH_INIT",
    SSH_MSG_KEX_DH_GEX_INIT           : "SSH_MSG_KEX_DH_GEX_INIT",
    SSH_MSG_KEX_DH_GEX_REPLY          : "SSH_MSG_KEX_DH_GEX_REPLY",
    SSH_MSG_KEX_DH_GEX_REQUEST        : "SSH_MSG_KEX_DH_GEX_REQUEST",
    SSH_MSG_USERAUTH_REQUEST          : "SSH_MSG_USERAUTH_REQUEST",
    SSH_MSG_USERAUTH_FAILURE          : "SSH_MSG_USERAUTH_FAILURE",
    SSH_MSG_USERAUTH_SUCCESS          : "SSH_MSG_USERAUTH_SUCCESS",
    SSH_MSG_USERAUTH_BANNER           : "SSH_MSG_USERAUTH_BANNER",
    SSH_MSG_GLOBAL_REQUEST            : "SSH_MSG_GLOBAL_REQUEST",
    SSH_MSG_CHANNEL_OPEN              : "SSH_MSG_CHANNEL_OPEN",
    SSH_MSG_CHANNEL_OPEN_CONFIRMATION : "SSH_MSG_CHANNEL_OPEN_CONFIRMATION",
    SSH_MSG_CHANNEL_WINDOW_ADJUST     : "SSH_MSG_CHANNEL_WINDOW_ADJUST",
    SSH_MSG_CHANNEL_DATA              : "SSH_MSG_CHANNEL_DATA",
    SSH_MSG_CHANNEL_EXTENDED_DATA     : "SSH_MSG_CHANNEL_EXTENDED_DATA",
    SSH_MSG_CHANNEL_EOF               : "SSH_MSG_CHANNEL_EOF",
    SSH_MSG_CHANNEL_CLOSE             : "SSH_MSG_CHANNEL_CLOSE",
    SSH_MSG_CHANNEL_REQUEST           : "SSH_MSG_CHANNEL_REQUEST",
    SSH_MSG_CHANNEL_SUCCESS           : "SSH_MSG_CHANNEL_SUCCESS",
}

recv_seqno=0
def my_recv_plain(s):
    global recv_seqno
    tmp_len=s.recv(4)
    if len(tmp_len)<4:
        print ("my_recv_plain() fatal error. incomplete read while reading packet len, got %d bytes" % len(tmp_len))
        hexdump.hexdump(tmp_len)
        exit(0)
    pkt_len=struct.unpack (">i", tmp_len)[0]
    if VERBOSITY>=2:
        print ("my_recv_plain() pkt_len: %d or 0x%x" % (pkt_len & 0xffffffff, pkt_len & 0xffffffff))
    tmp=s.recv(pkt_len)
    if VERBOSITY>=2:
        print ("my_recv_plain() got:")
        hexdump.hexdump(tmp_len+tmp)
    recv_seqno=recv_seqno+1
    return tmp_len+tmp

def my_recv_MAC(s):
    global recv_seqno
    tmp_len=s.recv(4)
    if len(tmp_len)<4:
        print ("my_recv_MAC() incomplete read while reading packet len, got only %d bytes" % len(tmp_len))
        hexdump.hexdump(tmp_len)
        exit(0)
    pkt_len=struct.unpack (">i", tmp_len)[0]
    if VERBOSITY>=2:
        print ("my_recv_MAC() pkt_len: %d or 0x%x" % (pkt_len & 0xffffffff, pkt_len & 0xffffffff))
    pkt_len=pkt_len+MAC_SIZE
    tmp=s.recv(pkt_len)
    if VERBOSITY>=2:
        print ("my_recv_MAC() got:")
        hexdump.hexdump(tmp_len+tmp)

    recv_seqno=recv_seqno+1
    return tmp_len+tmp

AES_BLOCK_SIZE_IN_BITS=128
AES_BLOCK_SIZE_IN_BYTES=AES_BLOCK_SIZE_IN_BITS//8

def my_recv_encrypted(s):
    global recv_seqno
    global serv_to_client_ctr
    if CIPHER=="aes128-ctr":
        KEY=key_D[0:128//8]
    elif CIPHER=="aes256-ctr":
        KEY=key_D
    else:
        assert False

    if VERBOSITY>=2:
        print ("my_recv_encrypted() begin")
    first_block=s.recv(AES_BLOCK_SIZE_IN_BYTES)
    if len(first_block)==0:
        print ("my_recv_encrypted() fatal error. recv() returned nothing")
        exit(0)
    if len(first_block)<AES_BLOCK_SIZE_IN_BYTES:
        print ("my_recv_encrypted() fatal error. incomplete read while reading packet len, got only %d bytes" % len(first_block))
        hexdump.hexdump(first_block)
        exit(0)

    ctr = Crypto.Util.Counter.new(128, little_endian=False, initial_value=serv_to_client_ctr)
    t = Crypto.Cipher.AES.new(KEY, Crypto.Cipher.AES.MODE_CTR, counter = ctr)
    first_block_decrypted = t.decrypt(first_block)
    if VERBOSITY>=2:
        print ("first block decrypted:")
        hexdump.hexdump(first_block_decrypted)
    pkt_len=struct.unpack(">i", first_block_decrypted[0:4])[0]
    if VERBOSITY>=2:
        print ("my_recv_encrypted() pkt_len: %d or 0x%x" % (pkt_len & 0xffffffff, pkt_len & 0xffffffff))

    already_read=AES_BLOCK_SIZE_IN_BYTES
    more_to_read=pkt_len-already_read+4
    rest=s.recv(more_to_read)

    ctr = Crypto.Util.Counter.new(128, little_endian=False, initial_value=serv_to_client_ctr)
    t = Crypto.Cipher.AES.new(KEY, Crypto.Cipher.AES.MODE_CTR, counter = ctr)
    all_decrypted = t.decrypt(first_block+rest)
    blocks_total=len(first_block+rest)//AES_BLOCK_SIZE_IN_BYTES
    if VERBOSITY>=2:
        print ("decrypted:")
        hexdump.hexdump(all_decrypted)
    serv_to_client_ctr=serv_to_client_ctr+blocks_total
    #print ("my_recv_encrypted() serv_to_client_ctr=0x%x" % serv_to_client_ctr)

    mac=s.recv(MAC_SIZE)
    if VERBOSITY>=2:
        print ("read MAC")
        hexdump.hexdump(mac)

    recv_seqno=recv_seqno+1
    return all_decrypted + mac

def my_recv(s, MAC=False):
    if MAC:
        if ENCRYPTION:
            return my_recv_encrypted(s)
        else:
            return my_recv_MAC(s)
    else:
        return my_recv_plain(s)

send_seqno=0
def my_send(s, buf):
    global send_seqno
    if VERBOSITY>=2:
        print ("my_send() sending:")
        hexdump.hexdump(buf)
    tmp=s.send(buf)
    send_seqno=send_seqno+1

# From http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-3.html (3.3)
def align2grain (i, grain):
    return (i + grain-1) & ~(grain-1)

def get_buf (buf, idx):
    length=struct.unpack (">i", buf[idx:idx+4])[0]
    s=buf[idx+4:idx+4+length]
    return s, idx+4+length

def get_str (buf, idx):
    s, l = get_buf(buf, idx)
    return s.decode("utf-8"), l

def get_mpint (buf, idx):
    s, idx = get_buf(buf, idx)
    i=int.from_bytes(s, byteorder='big')
    return i, idx

def get_u32 (buf, idx):
    rt=struct.unpack (">i", buf[idx:idx+4])[0]
    return rt, idx+4

def pack_buf(buf):
    return struct.pack(">i", len(buf)) + buf

def pack_str(s):
    return pack_buf(s.encode("utf-8"))

def binlog(x):
    rt=int(math.ceil(math.log(x, 2)))
    return rt

def bits_for_mpint(i):
    return int(math.ceil(binlog(i)))

def bytes_for_mpint(i):
    return bits_for_mpint(i)//8 + 1

# we add zero byte if most significant bit=1 and this number can be treated as negative
# but we must send number in unsigned form, so extend it
def cvt_to_mpint_and_add_zero_if_needed(i):
    buf=i.to_bytes(bytes_for_mpint(i), byteorder='big')
    if (buf[0]&0x80)==0x80:
        return b"\x00" + buf
    else:
        return buf

def pack_mpint(i):
    tmp=cvt_to_mpint_and_add_zero_if_needed(i)
    return pack_buf(tmp)

def unwrap_packet(buf):
    packet_len, padding_len = struct.unpack (">iB", buf[0:4+1])

    buf_with_padding=buf[4:]
    if packet_len!=len(buf_with_padding):
        print ("fatal error in unwrap_packet")
        print ("packet_len=", packet_len)
        print ("len(buf_with_padding)=", len(buf_with_padding))
        print ("exiting")
        exit(0)
    # remove padding
    payload=buf_with_padding[1:-padding_len]
    return payload

def remove_padding(buf):
    padding_len = struct.unpack (">B", buf[0:1])[0]
    return buf[1:-padding_len]

def unpack_serv_KEX(buf):
    if VERBOSITY>=1:
        print ("unpack_serv_KEX() begin")
    global from_serv_kex_algorithms
    global from_serv_server_host_algorithms
    global from_serv_encryption_algorithms
    global from_serv_mac_algorithms
    global from_serv_compression_algorithms
    global serv_KEX_blob

    serv_KEX=unwrap_packet(buf)
    idx=0
    if VERBOSITY>=2:
        print ("KEX from serv without len and padlen and msg_code:") # will be hashed as blob
        hexdump.hexdump(serv_KEX[1:])
    serv_KEX_blob=serv_KEX[1:]

    msg_code = struct.unpack (">B", serv_KEX[idx:idx+1])[0]
    idx=idx+1
    assert msg_code==SSH_MSG_KEXINIT
    cookie=serv_KEX[idx:idx+0x10]
    idx=idx+0x10
    #print ("cookie", cookie)

    s, idx = get_str(serv_KEX, idx)
    if VERBOSITY>=1:
        for t in s.split(","):
            print ("kex_algorithms:", t)
    from_serv_kex_algorithms=s

    s, idx = get_str(serv_KEX, idx)
    if VERBOSITY>=1:
        for t in s.split(","):
            print ("server_host_algorithms:", t)
    from_serv_server_host_algorithms=s

    s, idx = get_str(serv_KEX, idx)
    from_serv_encryption_algorithms=s

    s, idx = get_str(serv_KEX, idx)
    assert (s==from_serv_encryption_algorithms) # usually the same!

    if VERBOSITY>=1:
        for t in s.split(","):
            print ("encryption_algorithms:", t)

    s, idx = get_str(serv_KEX, idx)
    from_serv_mac_algorithms=s

    s, idx = get_str(serv_KEX, idx)
    assert s==from_serv_mac_algorithms

    if VERBOSITY>=1:
        for t in s.split(","):
            print ("mac_algorithms:", t)

    s, idx = get_str(serv_KEX, idx)
    from_serv_compression_algorithms=s

    s, idx = get_str(serv_KEX, idx)
    assert s==from_serv_compression_algorithms

    if VERBOSITY>=1:
        for t in s.split(","):
            print ("compression_algorithms:", t)

    s, idx = get_str(serv_KEX, idx)
    languages_client_to_server=s

    s, idx = get_str(serv_KEX, idx)
    assert s==languages_client_to_server
    if VERBOSITY>=1 and len(s)>0:
        for t in s.split(","):
            print ("languages:", t)

    first_KEX_packet_follows, reserverd = struct.unpack (">Bi", serv_KEX[idx:idx+1+4])
    assert first_KEX_packet_follows==0
    assert reserverd==0

def unpack_KEXDH_REPLY(buf):
    payload=unwrap_packet(buf)

    if VERBOSITY>=2:
        print ("unpack_KEXDH_REPLY():")
        hexdump.hexdump(payload)

    idx=0
    msg_code = struct.unpack (">B", payload[idx:idx+1])[0]
    assert msg_code==SSH_MSG_KEXDH_REPLY
    idx=idx+1
    p, idx = get_mpint(payload, idx)
    # one from /etc/ssh/moduli
    if VERBOSITY>=1:
        print ("DH GEX modulus (P):", hex(p))
        print ("binlog(P):", binlog(p))
    g, idx = get_mpint(payload, idx)
    if VERBOSITY>=1:
        print ("DH GEX base (G):", g)
    assert len(payload)==idx # be sure nothing else left in buffer
    return g, p

def wrap_packet(payload):
    if VERBOSITY>=3:
        print ("wrap_packet() payload:")
        hexdump.hexdump(payload)
    payload_len=len(payload)
    if VERBOSITY>=3:
        print ("wrap_packet() payload_len", payload_len)
    with_header=4+1+payload_len
    if ENCRYPTION==False:
        with_padding=align2grain(with_header, 8)
    else:
        with_padding=align2grain(with_header, 16)
    padlen=with_padding-with_header
    if padlen<4:
        if ENCRYPTION==False:
            padlen=padlen+8
        else:
            padlen=padlen+16
    if VERBOSITY>=3:
        print ("wrap_packet() padlen", padlen)

    padding=b'\xAA'*padlen
    payload_padded=struct.pack (">B", padlen) + payload + padding
    outbuf=struct.pack(">i", len(payload_padded)) + payload_padded
    return outbuf

def have_common_algs(list1, list2):
    set1=set(list1.split(","))
    set2=set(list2.split(","))
    return len(set1 & set2)!=0 # intersection: https://docs.python.org/3/library/stdtypes.html#frozenset.intersection

def pack_client_KEX():
    global client_KEX_blob
    if VERBOSITY>=2:
        print ("pack_client_KEX()")
    if EXT_INFO_C:
        kex_algorithms="diffie-hellman-group-exchange-sha256,ext-info-c"
    else:
        kex_algorithms="diffie-hellman-group-exchange-sha256" # doesn't work with crypto=none, dunno why

    #server_host_algorithms="ssh-rsa"
    #server_host_algorithms="rsa-sha2-256"
    server_host_algorithms="ssh-rsa,rsa-sha2-256,rsa-sha2-512"
    # also: rsa-sha2-512,rsa-sha2-256,ecdsa-sha2-nistp256,ssh-ed25519
    if ENCRYPTION:
        encryption_algorithms_client_to_server=CIPHER
    else:
        encryption_algorithms_client_to_server="none"
    encryption_algorithms_server_to_client=encryption_algorithms_client_to_server
    mac_algorithms_client_to_server=HASH
    mac_algorithms_server_to_client=mac_algorithms_client_to_server
    if ENCRYPTION==False:
        compression_algorithms_client_to_server="none,zlib@openssh.com"
    else:
        compression_algorithms_client_to_server="none" # doesn't work with crypto=none
    compression_algorithms_server_to_client=compression_algorithms_client_to_server
    languages_client_to_server=""
    languages_server_to_client=languages_client_to_server

    assert have_common_algs(kex_algorithms, from_serv_kex_algorithms)
    assert have_common_algs(server_host_algorithms, from_serv_server_host_algorithms)
    assert have_common_algs(encryption_algorithms_client_to_server, from_serv_encryption_algorithms)
    assert have_common_algs(mac_algorithms_client_to_server, from_serv_mac_algorithms)
    assert have_common_algs(compression_algorithms_client_to_server, from_serv_compression_algorithms)

    buf=struct.pack(">B", SSH_MSG_KEXINIT)
    cookie=b'\x11'*16
    buf=buf+cookie
    buf=buf+pack_str(kex_algorithms)
    buf=buf+pack_str(server_host_algorithms)
    buf=buf+pack_str(encryption_algorithms_client_to_server)
    buf=buf+pack_str(encryption_algorithms_server_to_client)
    buf=buf+pack_str(mac_algorithms_client_to_server)
    buf=buf+pack_str(mac_algorithms_server_to_client)
    buf=buf+pack_str(compression_algorithms_client_to_server)
    buf=buf+pack_str(compression_algorithms_server_to_client)
    buf=buf+pack_str(languages_client_to_server)
    buf=buf+pack_str(languages_server_to_client)
    first_KEX_packet_follows, reserverd = 0, 0
    buf=buf+struct.pack (">Bi", first_KEX_packet_follows, reserverd)

    if VERBOSITY>=2:
        print ("KEX from client without len and padlen and msg_code and padding:") # will be hashed as blob
        hexdump.hexdump(buf[1:])
    client_KEX_blob=buf[1:]

    return wrap_packet(buf)

def pack_DH_GEX_request():
    if VERBOSITY>=2:
        print ("pack_DH_GEX_request()")
    return wrap_packet(struct.pack(">Biii", SSH_MSG_KEX_DH_GEX_REQUEST, DH_GEX_min, DH_GEX_bits, DH_GEX_max))

def pack_DH_GEX_INIT(e):
    if VERBOSITY>=2:
        print ("pack_DH_GEX_INIT()")
    buf=struct.pack(">B", SSH_MSG_KEX_DH_GEX_INIT)
    buf=buf+pack_mpint(e)
    return wrap_packet(buf)

def pack_new_keys():
    if VERBOSITY>=2:
        print ("pack_new_keys()")
    buf=struct.pack(">B", SSH_MSG_NEWKEYS)
    return wrap_packet(buf)

def send_new_keys(s):
    tmp=pack_new_keys()
    if VERBOSITY>=2:
        print ("send_new_keys():")
        hexdump.hexdump(tmp)
    my_send(s, tmp)

def unpack_host_key(host_key):
    buf=host_key
    idx=0
    host_key_type, idx = get_str(buf, idx)
    RSA_e, idx = get_mpint(buf, idx)
    RSA_modulus_n, idx = get_mpint(buf, idx)
    assert len(buf)==idx # be sure nothing else left in buffer
    if VERBOSITY>=1:
        print ("host_key_type:", host_key_type)
        print ("RSA_e:", RSA_e)
        print ("RSA_modulus_n:", RSA_modulus_n)
        print ("binlog(RSA_modulus_n):", binlog(RSA_modulus_n))

def unpack_DH_GEX_REPLY(buf):
    idx=0
    if VERBOSITY>=2:
        print ("unpack_DH_GEX_REPLY():")
        hexdump.hexdump(buf)
    msg_code = struct.unpack (">B", buf[idx:idx+1])[0]
    if msg_code!=SSH_MSG_KEX_DH_GEX_REPLY:
        print ("expecting msg_code==SSH_MSG_KEX_DH_GEX_REPLY, but got", msg_code)
        exit(0)
    idx=idx+1
    KEX_host_key, idx = get_buf(buf, idx)
    if VERBOSITY>=2:
        print ("KEX_host_key: will be hashed as blob:")
        # same as in /etc/ssh/ssh_host_rsa_key.pub
        hexdump.hexdump(KEX_host_key)
    unpack_host_key(KEX_host_key)
    server_f, idx=get_mpint (buf, idx)
    if VERBOSITY>=1:
        print ("server_f: "+hex(server_f))
    KEX_H_sig, idx=get_buf (buf, idx) # TODO: handle this
    if VERBOSITY>=2:
        print ("KEX_H_sig:") # begins with the "rsa-sha2-256" string
        hexdump.hexdump(KEX_H_sig)
    assert len(buf)==idx # be sure nothing else left in buffer
    return KEX_host_key, server_f

# hash hodgepodge
# this is almost all data exchanged before this moment
# critical part - shared_secret
# [see kexgex_hash() in OpenSSH 9.0]
def calc_kexgex_hash(client_banner, serv_banner, client_KEX_blob, serv_KEX_blob, KEX_host_key, client_e, server_f, shared_secret, g, p):
    buf=b""
    buf=buf+pack_buf(client_banner)
    buf=buf+pack_buf(serv_banner)
    buf=buf+pack_buf(struct.pack(">B", SSH_MSG_KEXINIT)+client_KEX_blob)
    buf=buf+pack_buf(struct.pack(">B", SSH_MSG_KEXINIT)+serv_KEX_blob)
    buf=buf+pack_buf(KEX_host_key)
    buf=buf+struct.pack(">iii", DH_GEX_min, DH_GEX_bits, DH_GEX_max)
    buf=buf+pack_mpint(p)
    buf=buf+pack_mpint(g)
    buf=buf+pack_mpint(client_e)
    buf=buf+pack_mpint(server_f)
    buf=buf+pack_mpint(shared_secret)
    if VERBOSITY>=3:
        print ("buf to hash. len=", len(buf))
        hexdump.hexdump(buf)
    hashed=hashlib.sha256(buf).digest()
    if VERBOSITY>=2:
        print ("hashed:")
        #print (hashed)
        hexdump.hexdump(hashed)
    return hashed

def derive_key(K, H, c, session_id):
    m=hashlib.sha256()
    m.update(pack_mpint(K))
    m.update(H)
    m.update(str.encode(c))
    m.update(session_id)
    key=m.digest()
    if VERBOSITY>=2:
        print ("key_%c:" % c)
        hexdump.hexdump(key)
    return key

# [See kex.c, derive_key() in OpenSSH 9.0, also RFC 4253 7.2]
def derive_keys(K, H, session_id):
    global key_A
    global key_B
    global key_C
    global key_D
    global key_E
    global key_F
    global client_to_serv_ctr
    global serv_to_client_ctr
    if VERBOSITY>=2:
        print ("derive_keys() begin")
    key_A=derive_key(K, H, 'A', session_id)
    key_B=derive_key(K, H, 'B', session_id)
    key_C=derive_key(K, H, 'C', session_id)
    key_D=derive_key(K, H, 'D', session_id)
    key_E=derive_key(K, H, 'E', session_id)
    key_F=derive_key(K, H, 'F', session_id)
    # AES counters:
    client_to_serv_ctr = int.from_bytes(key_A[0:128//8], byteorder='big')
    serv_to_client_ctr = int.from_bytes(key_B[0:128//8], byteorder='big')

def my_send_with_MAC(s, msg):
    global client_to_serv_ctr
    global client_to_serv_ctr_1
    if CIPHER=="aes128-ctr":
        KEY=key_C[0:128//8]
    elif CIPHER=="aes256-ctr":
        KEY=key_C
    else:
        assert False
    if VERBOSITY>=2:
        print ("my_send_with_MAC() len(msg)=", len(msg))
        hexdump.hexdump(msg)
    calc_MAC = hmac.digest(key=key_E[0:MAC_SIZE], msg=struct.pack(">i", send_seqno-1)+msg, digest=MAC_ALGO)
    if ENCRYPTION==False:
        assert (len(msg)&0x7)==0
        my_send(s, msg+calc_MAC)
    else:
        assert (len(msg)&0xf)==0

        ctr = Crypto.Util.Counter.new(128, little_endian=False, initial_value=client_to_serv_ctr)
        t = Crypto.Cipher.AES.new(KEY, Crypto.Cipher.AES.MODE_CTR, counter = ctr)
        t2 = t.encrypt(msg)
        client_to_serv_ctr=client_to_serv_ctr+(len(t2)//AES_BLOCK_SIZE_IN_BYTES)
        my_send(s, t2+calc_MAC)
    if VERBOSITY>=2:
        print ("my_send_with_MAC() finish")

def my_recv_chk_MAC(s):
    tmp=my_recv(s, True)
    if VERBOSITY>=3:
        print ("my_recv_chk_MAC() begin. buf:")
        hexdump.hexdump(tmp)
    assert len(tmp)!=0
    tmplen=len(tmp)
    msg=tmp[0:-MAC_SIZE]
    msglen=len(msg)
    if VERBOSITY>=3:
        print ("msg to check:")
        hexdump.hexdump(msg)
    received_MAC=tmp[msglen:msglen+MAC_SIZE]
    assert len(received_MAC)==MAC_SIZE
    if VERBOSITY>=3:
        print ("my_recv_chk_MAC(): recv_seqno=", recv_seqno)

    calc_MAC = hmac.digest(key=key_F[0:MAC_SIZE], msg=struct.pack(">i", recv_seqno-2)+msg, digest=MAC_ALGO)

    if calc_MAC!=received_MAC:
        print ("fatal error: received MAC is not correct!")
        if VERBOSITY>=3:
            print ("calculated MAC:")
            hexdump.hexdump(calc_MAC)
        exit(0)
    else:
        if VERBOSITY>=3:
            print ("my_recv_chk_MAC(): MAC checked OK")
    return msg

def send_exec(s, cmd):
    print ("send_exec", cmd)
    msg=struct.pack("B", SSH_MSG_CHANNEL_REQUEST)
    msg=msg+struct.pack(">i", 0)
    msg=msg+pack_str("exec")
    msg=msg+struct.pack("B", 1)
    msg=msg+pack_str(cmd)
    pkt=wrap_packet(msg)
    #hexdump.hexdump(pkt)
    my_send_with_MAC(s, pkt)

# [see RFC4254 6.10]
def unpack_exit_status(msg):
    assert msg[0]==SSH_MSG_CHANNEL_REQUEST
    idx=1+4 # skip recipient channel
    tmp, idx = get_str(msg, idx)
    assert tmp=="exit-status"
    idx=idx+1 # skip boolean
    exit_status, idx = get_u32(msg, idx)
    print ("exit-status", exit_status)
    assert len(msg)==idx # be sure nothing else left in buffer

# [see RFC4254 5.2]
def unpack_channel_data(msg):
    assert msg[0]==SSH_MSG_CHANNEL_DATA
    idx=1+4 # skip recipient channel
    tmp, idx = get_str(msg, idx)
    print ("response:")
    print ("==")
    print (tmp)
    print ("==")
    assert len(msg)==idx # be sure nothing else left in buffer

# [see RFC4254 5.2]
SSH_EXTENDED_DATA_STDERR=1
def unpack_channel_extended_data(msg):
    assert msg[0]==SSH_MSG_CHANNEL_EXTENDED_DATA
    idx=1+4 # skip recipient channel
    code, idx = get_u32(msg, idx)
    assert code==SSH_EXTENDED_DATA_STDERR
    tmp, idx = get_str(msg, idx)
    print ("response (stderr):")
    print ("==")
    print (tmp)
    print ("==")
    assert len(msg)==idx # be sure nothing else left in buffer

# see [RFC4254 5.3]
def send_eof(s):
    if VERBOSITY>=2:
        print ("send_eof()")
    msg=struct.pack("B", SSH_MSG_CHANNEL_EOF)
    msg=msg+struct.pack(">i", 0)
    pkt=wrap_packet(msg)
    #hexdump.hexdump(pkt)
    my_send_with_MAC(s, pkt)

def send_channel_open(s):
    msg=struct.pack("B", SSH_MSG_CHANNEL_OPEN) # 90 or 0x5a
    msg=msg+pack_str("session")
    msg=msg+struct.pack(">iii", 0, 0x2000, 0x800000)
    pkt=wrap_packet(msg)
    #hexdump.hexdump(pkt)
    my_send_with_MAC(s, pkt)

def expect_msg(s, msg_code):
    if VERBOSITY>=2:
        print ("expecting msg_code=0x%x or:" % msg_code, msg_str[msg_code])
    msg=unwrap_packet(my_recv_chk_MAC(s))
    if msg[0]==msg_code:
        return msg
    print ("Fatal error. Got message 0x%x instead, or:" % msg[0], msg_str[msg[0]])
    exit(0)

def expect_new_keys(s):
    # plain message, unencrypted
    if VERBOSITY>=2:
        print ("expecting SSH_MSG_NEWKEYS")
    tmp=unwrap_packet(my_recv(s))
    if VERBOSITY>=2:
        print ("expect_new_keys(), got:")
        hexdump.hexdump(tmp)
    assert tmp==struct.pack("B", SSH_MSG_NEWKEYS)

def send_ssh_userauth(s):
    if VERBOSITY>=1:
        print ("sending ssh-userauth")
    msg=wrap_packet(struct.pack("B", SSH_MSG_SERVICE_REQUEST)+pack_str("ssh-userauth"))
    if VERBOSITY>=2:
        print ("send_ssh_userauth(), sending:")
        hexdump.hexdump(msg)
    my_send_with_MAC(s, msg)

def send_ssh_connection_none(s):
    global USERNAME
    if VERBOSITY>=1:
        print ("sending ssh-connection")
    msg=struct.pack("B", SSH_MSG_USERAUTH_REQUEST)
    if USERNAME==None:
        USERNAME="root"
    msg=msg+pack_str(USERNAME)
    msg=msg+pack_str("ssh-connection")
    msg=msg+pack_str("none")
    pkt=wrap_packet(msg)
    if VERBOSITY>=2:
        print ("send_ssh_connection_none(), sending:")
        hexdump.hexdump(pkt)
    my_send_with_MAC(s, pkt)

def send_ssh_connection_password(s, username, password):
    if VERBOSITY>=1:
        print ("sending ssh-connection")
    msg=struct.pack("B", SSH_MSG_USERAUTH_REQUEST)
    msg=msg+pack_str(username)
    msg=msg+pack_str("ssh-connection")
    msg=msg+pack_str("password")
    msg=msg+b"\x00" # change password byte
    msg=msg+pack_str(password)
    pkt=wrap_packet(msg)
    if VERBOSITY>=2:
        print ("send_ssh_connection_password(), sending:")
        hexdump.hexdump(pkt)
    my_send_with_MAC(s, pkt)
    msg=unwrap_packet(my_recv_chk_MAC(s))
    if msg[0]==SSH_MSG_USERAUTH_SUCCESS:
        print ("Password correct.")
        return True
    elif msg[0]==SSH_MSG_USERAUTH_FAILURE:
        print ("Fatal error. Password incorrect.")
        return False
    else:
        assert False

def get_list_of_auth_modes(s):
    # this message sent to get list of auth modes:
    send_ssh_connection_none(s)
    msg=expect_msg(s, SSH_MSG_USERAUTH_FAILURE)
    tmp, _ = get_str(msg, 1)
    if VERBOSITY>=1:
        print ("allowed auth modes:", tmp)
    return tmp

def send_client_KEX(s):
    tmp=pack_client_KEX()
    if VERBOSITY>=2:
        print ("client KEX:")
        hexdump.hexdump(tmp)
    my_send(s, tmp)

def DH_get_params(s):
    my_send(s, pack_DH_GEX_request())
    tmp=my_recv(s)
    assert len(tmp)!=0
    g, p = unpack_KEXDH_REPLY(tmp)
    return g, p

def DH_send_client_e(s, g, p):
    # ATTENTION: may be insecure. use better PRNG
    my_secret=random.randrange(1, 2**DH_GEX_bits)
    client_e=pow(g, my_secret, p)

    if VERBOSITY>=1:
        print ("client_e", hex(client_e))

    my_send(s, pack_DH_GEX_INIT(client_e))
    return my_secret, client_e

def DH_recv_server_f(s):
    tmp=my_recv(s)
    assert len(tmp)!=0
    KEX_host_key, server_f=unpack_DH_GEX_REPLY(unwrap_packet(tmp))
    return KEX_host_key, server_f

def DH_do_exchange(s):
    g, p = DH_get_params(s)

    my_secret, client_e=DH_send_client_e(s, g, p)

    KEX_host_key, server_f = DH_recv_server_f(s)

    # calculate DH shared secret:
    shared_secret=pow(server_f, my_secret, p)
    if VERBOSITY>=1:
        print ("shared_secret: "+hex(shared_secret))
    return g, p, KEX_host_key, client_e, server_f, shared_secret

def expect_service_accept(s):
    # waiting for SSH_MSG_SERVICE_ACCEPT
    while True:
        msg=unwrap_packet(my_recv_chk_MAC(s))
        #hexdump.hexdump(msg)
        if msg[0]==SSH_MSG_SERVICE_ACCEPT:
            break
        if msg[0]==SSH_MSG_IGNORE:
            # "markus" packet, sent by server if DEBUG_KEXDH is on [see sshd.c in OpenSSH 9.0]
            # it is the first encrypted/MACed packed server sends as a test
            # to be ignored
            if VERBOSITY>=2:
                print ("got 'markus' packet:")
                hexdump.hexdump(msg)
        if msg[0]!=SSH_MSG_IGNORE:
            print ("error, unhandled packet type at this stage:", msg[0], msg_str[msg[0]])

def login(s):
    serv_auth_modes=get_list_of_auth_modes(s)
    if "password" not in serv_auth_modes:
        print ("Fatal error: server doesn't accept password auth")
        exit(0)

    if USERNAME==None or PASSWORD==None:
        print ("we can't go futher without username/passsword. exiting.")
        exit(0)
    if send_ssh_connection_password(s, USERNAME, PASSWORD)==False:
        exit(0)

def exec_recv_response(s):
    # receiving loop
    # as in https://en.wikipedia.org/wiki/Inversion_of_control
    # as in GUI
    while True:
        msg=unwrap_packet(my_recv_chk_MAC(s))
        msg_code=msg[0]
        if VERBOSITY>=1:
            print ("got message 0x%x or %s" % (msg_code, msg_str[msg_code]))
        if msg_code==SSH_MSG_CHANNEL_WINDOW_ADJUST:
            # see [RFC4254 5.2]
            pass
        elif msg_code==SSH_MSG_CHANNEL_SUCCESS:
            pass
        elif msg_code==SSH_MSG_CHANNEL_DATA:
            unpack_channel_data(msg)
        elif msg_code==SSH_MSG_CHANNEL_EXTENDED_DATA:
            unpack_channel_extended_data(msg)
        elif msg_code==SSH_MSG_CHANNEL_EOF:
            pass
        elif msg_code==SSH_MSG_CHANNEL_REQUEST:
            unpack_exit_status(msg)
        elif msg_code==SSH_MSG_CHANNEL_CLOSE:
            break
        else:
            print ("Fatal error. Unhandled msg_code=0x%x or %s" % (msg_code, msg_str[msg_code]))
            exit(0)

def exec_command(s, cmd):
    send_exec(s, cmd)
    exec_recv_response(s)

def do_all():
    global recv_seqno
    
    s = socket.socket(socket.AF_INET)
    s.connect((HOST, PORT))

    tmp=s.recv(1024)
    if VERBOSITY>=2:
        print ("got from server:")
        hexdump.hexdump(tmp)
    serv_banner=tmp.rstrip()
    if VERBOSITY>=1:
        print ("serv_banner:", serv_banner.decode("utf-8"))
    recv_seqno=recv_seqno+1

    my_send(s, client_banner+b"\r\n")

    unpack_serv_KEX(my_recv(s))

    send_client_KEX(s)

    g, p, KEX_host_key, client_e, server_f, shared_secret = DH_do_exchange(s)

    # calculate the hash of hodgepodge. the most important part - shared_secret, which is unknown to interceptor
    # all other parts can be easily intercepted!
    kexgex_hash=calc_kexgex_hash(client_banner, serv_banner, client_KEX_blob, serv_KEX_blob, KEX_host_key, client_e, server_f, shared_secret, g, p)
    if VERBOSITY>=2:
        print ("kexgex_hash:")
        hexdump.hexdump(kexgex_hash)

    derive_keys(shared_secret, kexgex_hash, kexgex_hash)

    expect_new_keys(s)

    if EXT_INFO_C:
        expect_msg(s, SSH2_MSG_EXT_INFO)

    send_new_keys(s)

    send_ssh_userauth(s)

    expect_service_accept(s)

    login(s)

    send_channel_open(s)

    # [see RFC4254 4]: https://www.rfc-editor.org/rfc/rfc4254#section-4
    expect_msg(s, SSH_MSG_GLOBAL_REQUEST)

    expect_msg(s, SSH_MSG_CHANNEL_OPEN_CONFIRMATION)

    exec_command(s, COMMAND)

    # be polite
    send_eof(s)
    s.close()

parse_command_line()
if HOST==None:
    print ("Fatal error: host not set")
    exit(0)
do_all()

